import time

from django.contrib import messages
from django.contrib.auth import logout
from portal.app_settings import SESSION_EXPIRY_TIME


class SessionTimeoutMiddleware:
    """
    Logs the user out after a predefined time of inactivity.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.user.is_authenticated:
            if "last_request" in request.session:
                elapsed_seconds = time.time() - request.session["last_request"]
                if elapsed_seconds > SESSION_EXPIRY_TIME:
                    del request.session["last_request"]
                    logout(request)
                    messages.info(request, "You have been logged out due to inactivity.")
            request.session["last_request"] = time.time()
        else:
            if "last_request" in request.session:
                del request.session["last_request"]

        response = self.get_response(request)

        return response



from portal.tests.pageObjects.portal.base_page import BasePage


class ForbiddenPage(BasePage):
    def __init__(self, browser):
        super(ForbiddenPage, self).__init__(browser)
        assert self.on_correct_page("403_forbidden")



RAPID_ROUTER_RESOURCES_BANNER = {
    "title": "Rapid Router Resources",
    "subtitle": "We’ve created a comprehensive set of teaching materials to help you "
    "teach students Computing.",
    "text": "",
    "image_class": "banner--picture--play",
    "image_description": "Credit: Annie Spratt, Unsplash",
    "alt": "Boy playing on ipad",
}



"""
This migration essentially "replaces" the majority of the models in models.py with
the models in common/models.py.
The models affected here are UserProfile, School, Teacher, Class and Student.
For each of these models, this migration:
- removes each of the model fields,
- updates Guardian's foreign keys to use the models in common/models.py,
- deletes the model, and updates the DB's references of these models to the new ones
created in common/models.py.

The important thing to note here is the use of migrations.SeparateDatabaseAndState.
This operation makes it possible to makes different changes to the state and to the
database. Essentially this is used here to replace Portal's models with Common's
models, without moving or deleting anything from the database.
This migration has been made following the tutorial on how to move Django models:
https://realpython.com/move-django-model/ (following the Django way example).

Since Rapid Router's models used to reference Portal's models, Rapid Router's migration
0071 needs to happen first, to avoid causing an error when migrating.
"""
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ("common", "0001_initial"),
        ("game", "0071_use_common_models"),
        ("portal", "0057_delete_frontpagenews"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="class",
                    name="teacher",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="student",
                    name="class_field",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="student",
                    name="new_user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="student",
                    name="pending_class_request",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="student",
                    name="user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="teacher",
                    name="new_user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="teacher",
                    name="pending_join_request",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="teacher",
                    name="school",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="teacher",
                    name="user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="userprofile",
                    name="user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name="guardian",
                    name="children",
                    field=models.ManyToManyField(to="common.Student"),
                ),
            ],
            # You're reusing an existing table, so do nothing
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name="guardian",
                    name="user",
                    field=models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="common.UserProfile",
                    ),
                ),
            ],
            # You're reusing an existing table, so do nothing
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Class",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="Class",
                    table="common_class",
                ),
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="School",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="School",
                    table="common_school",
                ),
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Student",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="Student",
                    table="common_student",
                ),
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="Teacher",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="Teacher",
                    table="common_teacher",
                ),
            ],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="UserProfile",
                ),
            ],
            database_operations=[
                migrations.AlterModelTable(
                    name="UserProfile",
                    table="common_userprofile",
                ),
            ],
        ),
    ]



from __future__ import print_function

import os
import time
from builtins import str

from hamcrest import assert_that, equal_to, contains_string
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.base_page import BasePage


class GamePage(BasePage):
    def __init__(self, browser):
        super(GamePage, self).__init__(browser)

        assert self.on_correct_page("game_page")

        self._dismiss_initial_dialog()

    def _dismiss_initial_dialog(self):
        self.dismiss_dialog("play_button")
        return self

    def dismiss_dialog(self, button_id):
        self.wait_for_element_to_be_clickable((By.ID, button_id))
        self.browser.find_element(By.ID, button_id).click()
        self.wait_for_element_to_be_invisible((By.ID, button_id))

    def load_solution(self, workspace_id):
        self.browser.find_element(By.ID, "load_tab").click()
        selector = "#loadWorkspaceTable tr[value='" + str(workspace_id) + "']"
        self.wait_for_element_to_be_clickable((By.CSS_SELECTOR, selector))
        self.browser.find_element(By.CSS_SELECTOR, selector).click()
        self.browser.find_element(By.ID, "loadWorkspace").click()
        time.sleep(1)
        return self

    def clear(self):
        self.browser.find_element(By.ID, "clear_tab").click()
        return self

    def try_again(self):
        self.dismiss_dialog("try_again_button")
        return self

    def step(self):
        self.browser.find_element(By.ID, "step_tab").click()
        return self

    def assert_is_green_light(self, traffic_light_index):
        self._assert_light_is_on(traffic_light_index, "green")

    def assert_is_red_light(self, traffic_light_index):
        self._assert_light_is_on(traffic_light_index, "red")

    def _assert_light_is_on(self, traffic_light_index, colour):
        image = self.browser.find_element(By.ID, "trafficLight_%s_%s" % (traffic_light_index, colour))

        assert_that(image.get_attribute("opacity"), equal_to("1"))

    def run_program(self, wait_for_element_id="algorithmScore"):
        self.browser.find_element(By.ID, "fast_tab").click()

        try:
            self.wait_for_element_to_be_clickable((By.ID, wait_for_element_id), 45)
        except TimeoutException as e:
            import time

            millis = int(round(time.time() * 1000))
            screenshot_filename = "/tmp/game_tests_%s-%s.png" % (os.getenv("BUILD_NUMBER", "nonumber"), str(millis))
            print("Saved screenshot to " + screenshot_filename)
            self.browser.get_screenshot_as_file(screenshot_filename)
            raise e

        return self

    def run_crashing_program(self):
        return self._run_failing_program("What went wrong")

    def run_program_that_runs_out_of_instructions(self):
        return self._run_failing_program("The van ran out of instructions before it reached a destination.")

    def _run_failing_program(self, text):
        self.run_program("try_again_button")
        error_message = self.browser.find_element(By.ID, "myModal-lead").text
        assert_that(error_message, contains_string(text))
        return self

    def _assert_score(self, element_id, score):
        route_score = self.browser.find_element(By.ID, element_id).text
        assert_that(route_score, equal_to(score))
        return self

    def assert_route_score(self, score):
        return self._assert_score("routeScore", score)

    def assert_algorithm_score(self, score):
        return self._assert_score("algorithmScore", score)



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from . import class_page, onboarding_student_list_page
from .teach_base_page import TeachBasePage


class EditStudentPage(TeachBasePage):
    def __init__(self, browser):
        super(EditStudentPage, self).__init__(browser)

        assert self.on_correct_page("edit_student_page")

    def type_student_name(self, name):
        self.browser.find_element(By.ID, "id_name").clear()
        self.browser.find_element(By.ID, "id_name").send_keys(name)
        return self

    def type_student_password(self, password):
        self.browser.find_element(By.ID, "id_password").send_keys(password)
        self.browser.find_element(By.ID, "id_confirm_password").send_keys(password)
        return self

    def click_update_button(self):
        self.browser.find_element(By.ID, "update_name_button").click()
        return class_page.TeachClassPage(self.browser)

    def click_update_button_fail(self):
        self.browser.find_element(By.ID, "update_name_button").click()
        return self

    def click_set_password_button(self):
        self.browser.find_element(By.ID, "set_new_password_button").click()
        return onboarding_student_list_page.OnboardingStudentListPage(self.browser)

    def is_student_name(self, name):
        return name in self.browser.find_element(By.ID, "student_details").text



from __future__ import absolute_import

from .base_page import BasePage


class ResourcesPage(BasePage):
    def __init__(self, browser):
        super(ResourcesPage, self).__init__(browser)

        assert self.on_correct_page("resources_page")



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0046_auto_20150723_1101")]

    operations = [migrations.RemoveField(model_name="userprofile", name="avatar")]



"""
This file is a customisation of the django-ratelimit library. The purpose of this
customisation is to give us the ability to access and the delete the cache key
associated with the ratelimit counter. We want this so that we can reset the
ratelimit counter whenever a successful request is performed on any ratelimited
view.

is_ratelimited() and get_usage() are the same as in django-ratelimit, with the
exception of the line `global cache_key` at the start of get_usage().
is_ratelimited() is called in the customised django-ratelimit ratelimit
decorator found in portal/helpers/decorators.py.

`get_ratelimit_cache_key_for_user`, `get_ratelimit_count_for_user`
and `clear_ratelimit_cache_for_user` are custom functions.

More info on the core methods of django-ratelimit can be found here:
https://django-ratelimit.readthedocs.io/en/stable/usage.html#core-methods
"""
from __future__ import absolute_import

import functools
import time

from django.conf import settings
from django.core.cache import caches, cache
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from ratelimit import ALL, UNSAFE
from ratelimit.core import (
    _split_rate,
    _method_match,
    _ACCESSOR_KEYS,
    _SIMPLE_KEYS,
    _get_window,
    EXPIRATION_FUDGE,
    _make_cache_key,
)

from portal.helpers.regexes import EMAIL_REGEX_PATTERN
from portal.helpers.request_handlers import get_access_code_from_request

RATELIMIT_LOGIN_GROUP = "login"
RATELIMIT_LOGIN_RATE = "5/d"
RATELIMIT_LOGIN_RATE_SCHOOL_STUDENT = "10/d"
RATELIMIT_METHOD = "POST"

RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_GROUP = "user_already_registered_email"
RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_RATE = "1/d"


def school_student_key(group, request):
    access_code = get_access_code_from_request(request)
    return f'{request.POST.get("username", "")},{access_code}'


def get_ratelimit_cache_key_for_user(user: str):
    # check for email quickly

    user_rate = RATELIMIT_LOGIN_RATE if EMAIL_REGEX_PATTERN.match(user) else RATELIMIT_LOGIN_RATE_SCHOOL_STUDENT
    _, period = _split_rate(rate=user_rate)
    window = _get_window(value=user, period=period)
    cache_key = _make_cache_key(
        group=RATELIMIT_LOGIN_GROUP,
        window=window,
        rate=user_rate,
        value=user,
        methods=RATELIMIT_METHOD,
    )
    return cache_key


def get_ratelimit_count_for_user(user: str):
    cache_key = get_ratelimit_cache_key_for_user(user)
    return cache.get(cache_key)


def clear_ratelimit_cache_for_user(user: str):
    cache_key = get_ratelimit_cache_key_for_user(user)
    cache.delete(cache_key)


def is_ratelimited(request, group=None, fn=None, key=None, rate=None, method=ALL, increment=False):
    """
    As in django-ratelimit. Calls "get_usage" defined below to enable the usage of
    the custom cache_key functionality.
    """
    usage = get_usage(request, group, fn, key, rate, method, increment)
    if usage is None:
        return False

    return usage["should_limit"]


def get_usage(request, group=None, fn=None, key=None, rate=None, method=ALL, increment=False):
    """
    As in django-ratelimit. Makes cache_key global so it can be called outside the scope
    and the cache can be accessed at later times.
    """
    global cache_key

    if group is None and fn is None:
        raise ImproperlyConfigured("get_usage must be called with either " "`group` or `fn` arguments")

    if not getattr(settings, "RATELIMIT_ENABLE", True):
        return None

    if not _method_match(request, method):
        return None

    if group is None:
        parts = []

        if isinstance(fn, functools.partial):
            fn = fn.func

        # Django <2.1 doesn't use a partial. This is ugly and inelegant, but
        # throwing __qualname__ into the list below helps.
        if fn.__name__ == "bound_func":
            fn = fn.__closure__[0].cell_contents

        if hasattr(fn, "__module__"):
            parts.append(fn.__module__)

        if hasattr(fn, "__self__"):
            parts.append(fn.__self__.__class__.__name__)

        parts.append(fn.__qualname__)
        group = ".".join(parts)

    if callable(rate):
        rate = rate(group, request)
    elif isinstance(rate, str) and "." in rate:
        ratefn = import_string(rate)
        rate = ratefn(group, request)

    if rate is None:
        return None
    limit, period = _split_rate(rate)
    if period <= 0:
        raise ImproperlyConfigured("Ratelimit period must be greater than 0")

    if not key:
        raise ImproperlyConfigured("Ratelimit key must be specified")
    if callable(key):
        value = key(group, request)
    elif key in _SIMPLE_KEYS:
        value = _SIMPLE_KEYS[key](request)
    elif ":" in key:
        accessor, k = key.split(":", 1)
        if accessor not in _ACCESSOR_KEYS:
            raise ImproperlyConfigured("Unknown ratelimit key: %s" % key)
        value = _ACCESSOR_KEYS[accessor](request, k)
    elif "." in key:
        keyfn = import_string(key)
        value = keyfn(group, request)
    else:
        raise ImproperlyConfigured("Could not understand ratelimit key: %s" % key)

    window = _get_window(value, period)
    initial_value = 1 if increment else 0

    cache_name = getattr(settings, "RATELIMIT_USE_CACHE", "default")
    cache = caches[cache_name]

    if value == "":
        return {
            "count": 0,
            "limit": 0,
            "should_limit": False,
            "time_left": -1,
        }
    else:
        cache_key = _make_cache_key(group, window, rate, value, method)

    count = None
    added = cache.add(cache_key, initial_value, period + EXPIRATION_FUDGE)
    if added:
        count = initial_value
    else:
        if increment:
            try:
                # python3-memcached will throw a ValueError if the server is
                # unavailable or (somehow) the key doesn't exist. redis, on the
                # other hand, simply returns None.
                count = cache.incr(cache_key)
            except ValueError:
                pass
        else:
            count = cache.get(cache_key, initial_value)

    # Getting or setting the count from the cache failed
    if count is None:
        if getattr(settings, "RATELIMIT_FAIL_OPEN", False):
            return None
        return {
            "count": 0,
            "limit": 0,
            "should_limit": True,
            "time_left": -1,
        }

    time_left = window - int(time.time())
    return {
        "count": count,
        "limit": limit,
        "should_limit": count > limit,
        "time_left": time_left,
    }


is_ratelimited.ALL = ALL
is_ratelimited.UNSAFE = UNSAFE
get_usage.ALL = ALL
get_usage.UNSAFE = UNSAFE



from django.shortcuts import render

from portal.strings.about import ABOUT_BANNER, GETINVOLVED_BANNER, CONTRIBUTE_BANNER


def about(request):
    return render(request, "portal/about.html", {"BANNER": ABOUT_BANNER})


def getinvolved(request):
    return render(request, "portal/getinvolved.html", {"BANNER": GETINVOLVED_BANNER})


def contribute(request):
    return render(request, "portal/contribute.html", {"BANNER": CONTRIBUTE_BANNER})



import datetime

import pytest
from common.helpers.emails import (
    add_consent_record_to_dotmailer_user,
    add_contact_to_address_book,
    add_to_dotmailer,
    create_contact,
    delete_contact,
    send_dotmailer_consent_confirmation_email_to_user,
    DotmailerUserType,
)
from common.mail import address_book_ids
from django.test import Client
from django.urls import reverse

from example_project.portal_test_settings import (
    DOTMAILER_USER,
    DOTMAILER_PASSWORD,
    DOTMAILER_SEND_CAMPAIGN_URL,
    DOTMAILER_PUT_CONSENT_DATA_URL,
    DOTMAILER_THANKS_FOR_STAYING_CAMPAIGN_ID,
    DOTMAILER_CREATE_CONTACT_URL,
    DOTMAILER_DELETE_USER_BY_ID_URL,
    DOTMAILER_TEACHER_ADDRESS_BOOK_URL,
    DOTMAILER_STUDENT_ADDRESS_BOOK_URL,
    DOTMAILER_NO_ACCOUNT_ADDRESS_BOOK_URL,
)

FAKE_TIME = datetime.datetime(2020, 12, 25, 17, 5, 55)


@pytest.fixture
def patch_datetime_now(monkeypatch):
    class mydatetime:
        @classmethod
        def now(cls):
            return FAKE_TIME

    monkeypatch.setattr(datetime, "datetime", mydatetime)


def test_newsletter_calls_correct_requests(mocker, monkeypatch):
    mocked_create_contact = mocker.patch("common.helpers.emails.create_contact")
    mocked_add_to_address_book = mocker.patch(
        "common.helpers.emails.add_contact_to_address_book"
    )

    add_to_dotmailer(
        "Ray",
        "Charles",
        "ray.charles@example.com",
        address_book_ids["newsletter"],
        DotmailerUserType.TEACHER,
    )

    mocked_create_contact.assert_called_once()
    mocked_add_to_address_book.assert_called_once()


@pytest.mark.django_db
def test_newsletter_get_not_allowed():
    c = Client()

    response = c.get(reverse("process_newsletter_form"))

    assert response.status_code == 405


def test_delete_account(mocker):
    mocked_delete = mocker.patch("common.helpers.emails.delete")
    mocker.patch("common.helpers.emails.get_dotmailer_user_by_email")

    delete_contact("example@mail.com")

    mocked_delete.assert_called_once_with(
        DOTMAILER_DELETE_USER_BY_ID_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
    )


def test_newsletter_sends_correct_request_data(
    mocker, monkeypatch, patch_datetime_now
):
    mocked_post = mocker.patch("common.helpers.emails.post")

    expected_body1 = {
        "contact": {
            "email": "ray.charles@example.com",
            "optInType": "VerifiedDouble",
            "emailType": "Html",
            "dataFields": [
                {"key": "FIRSTNAME", "value": "Ray"},
                {"key": "LASTNAME", "value": "Charles"},
                {"key": "FULLNAME", "value": "Ray Charles"},
            ],
        },
        "consentFields": [
            {
                "fields": [
                    {"key": "DATETIMECONSENTED", "value": FAKE_TIME.__str__()}
                ]
            }
        ],
        "preferences": [{"trout": True}],
    }

    expected_body2 = {
        "email": "ray.charles@example.com",
        "optInType": "VerifiedDouble",
        "emailType": "Html",
        "dataFields": [
            {"key": "FIRSTNAME", "value": "Ray"},
            {"key": "LASTNAME", "value": "Charles"},
            {"key": "FULLNAME", "value": "Ray Charles"},
        ],
    }

    create_contact("Ray", "Charles", "ray.charles@example.com")

    mocked_post.assert_called_once_with(
        DOTMAILER_CREATE_CONTACT_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body1,
    )

    add_contact_to_address_book(
        "Ray",
        "Charles",
        "ray.charles@example.com",
        address_book_ids["newsletter"],
        DotmailerUserType.TEACHER,
    )

    assert mocked_post.call_count == 3

    mocked_post.assert_any_call(
        DOTMAILER_TEACHER_ADDRESS_BOOK_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body2,
    )

    mocked_post.reset_mock()

    add_contact_to_address_book(
        "Ray",
        "Charles",
        "ray.charles@example.com",
        address_book_ids["newsletter"],
        DotmailerUserType.STUDENT,
    )

    assert mocked_post.call_count == 2

    mocked_post.assert_any_call(
        DOTMAILER_STUDENT_ADDRESS_BOOK_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body2,
    )

    mocked_post.reset_mock()

    add_contact_to_address_book(
        "Ray",
        "Charles",
        "ray.charles@example.com",
        address_book_ids["newsletter"],
        DotmailerUserType.NO_ACCOUNT,
    )

    assert mocked_post.call_count == 2

    mocked_post.assert_any_call(
        DOTMAILER_NO_ACCOUNT_ADDRESS_BOOK_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body2,
    )


def test_consent_calls_send_correct_request_data(
    mocker, monkeypatch, patch_datetime_now
):
    mocked_post = mocker.patch("common.helpers.emails.post")
    mocked_put = mocker.patch("common.helpers.emails.put")

    user = {
        "id": 1,
        "email": "ray.charles@example.com",
        "optInType": "VerifiedDouble",
        "emailType": "Html",
        "dataFields": [
            {"key": "FIRSTNAME", "value": "Ray"},
            {"key": "LASTNAME", "value": "Charles"},
            {"key": "FULLNAME", "value": "Ray Charles"},
        ],
        "status": "Subscribed",
    }

    expected_body1 = {
        "contact": {
            "email": "ray.charles@example.com",
            "optInType": "VerifiedDouble",
            "emailType": "Html",
            "dataFields": [
                {"key": "FIRSTNAME", "value": "Ray"},
                {"key": "LASTNAME", "value": "Charles"},
                {"key": "FULLNAME", "value": "Ray Charles"},
            ],
        },
        "consentFields": [
            {
                "fields": [
                    {"key": "DATETIMECONSENTED", "value": FAKE_TIME.__str__()}
                ]
            }
        ],
    }

    expected_body2 = {
        "campaignID": DOTMAILER_THANKS_FOR_STAYING_CAMPAIGN_ID,
        "contactIds": ["1"],
    }

    add_consent_record_to_dotmailer_user(user)

    mocked_put.assert_called_once_with(
        DOTMAILER_PUT_CONSENT_DATA_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body1,
    )

    send_dotmailer_consent_confirmation_email_to_user(user)

    mocked_post.assert_called_with(
        DOTMAILER_SEND_CAMPAIGN_URL,
        auth=(DOTMAILER_USER, DOTMAILER_PASSWORD),
        json=expected_body2,
    )


@pytest.mark.django_db
def test_dotmailer_consent_form(mocker, monkeypatch):
    """
    Checks the various success / failures conditions of the Dotmailer consent form. The form contains two widgets,
    an email input field and a consent checkbox. The cases are as follows:
    - (invalid email format, consent given) -> form is invalid, error message is shown
    - (valid email format, consent not given) -> form is invalid, error message is shown
    - (valid email format, consent given) -> form is valid, user is redirected to home page
    - (non-existent email - Key Error, consent given) -> form is invalid, error is message is shown
    """
    c = Client()
    consent_form_url = reverse("consent_form")

    mocked_get_user_success = mocker.patch(
        "portal.views.dotmailer.get_dotmailer_user_by_email"
    )
    mocked_add_consent = mocker.patch(
        "portal.views.dotmailer.add_consent_record_to_dotmailer_user"
    )
    mocked_send_campaign = mocker.patch(
        "portal.views.dotmailer.send_dotmailer_consent_confirmation_email_to_user"
    )

    get_consent_form_response = c.get(consent_form_url)

    assert get_consent_form_response.status_code == 200

    bad_email_data = {"email": "fakeemail", "consent_ticked": True}
    no_consent_data = {"email": "real@email.com", "consent_ticked": False}
    good_request_data = {"email": "real@email.com", "consent_ticked": True}

    bad_email_response = c.post(consent_form_url, data=bad_email_data)

    assert bad_email_response.status_code == 302
    assert bad_email_response.url == consent_form_url
    _is_warning_message_showing(bad_email_response)

    no_consent_response = c.post(consent_form_url, data=no_consent_data)

    assert no_consent_response.status_code == 302
    assert no_consent_response.url == consent_form_url
    _is_warning_message_showing(no_consent_response)

    good_request_response = c.post(consent_form_url, data=good_request_data)

    assert good_request_response.status_code == 302
    assert good_request_response.url == reverse("home")
    mocked_get_user_success.assert_called_once()
    mocked_add_consent.assert_called_once()
    mocked_send_campaign.assert_called_once()

    mocker.patch(
        "portal.views.dotmailer.add_consent_record_to_dotmailer_user",
        side_effect=KeyError,
    )

    wrong_email_response = c.post(consent_form_url, data=good_request_data)

    assert wrong_email_response.status_code == 302
    assert wrong_email_response.url == consent_form_url
    _is_warning_message_showing(wrong_email_response)


def _is_warning_message_showing(response):
    messages = list(response.wsgi_request._messages)
    assert (
        messages[0].message
        == "Valid email address and consent required. Please try again."
    )



from __future__ import absolute_import

from builtins import str

from common.models import School, Student, UserProfile
from common.tests.utils.classes import create_class_directly
from common.tests.utils.teacher import signup_teacher_directly
from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse, reverse_lazy


class SecurityTestCase(TestCase):
    def _test_incorrect_teacher_cannot_login(self, view_name):
        email1, _ = signup_teacher_directly()
        email2, pass2 = signup_teacher_directly()
        _, _, access_code = create_class_directly(email1)

        c = Client()
        assert c.login(username=email2, password=pass2)
        page = reverse(view_name, args=[access_code])
        assert not c.get(page).status_code == 200

    def _test_incorrect_teacher_no_info_leak(self, view_name):
        email1, _ = signup_teacher_directly()
        email2, pass2 = signup_teacher_directly()
        _, _, access_code = create_class_directly(email1)

        c = Client()
        assert c.login(username=email2, password=pass2)

        invalid_page = reverse(view_name, args=[access_code])
        invalid_login_code = c.get(invalid_page).status_code

        non_existent_page = reverse(view_name, args=["AAAAA"])
        non_existent_code = c.get(non_existent_page).status_code

        assert non_existent_code == invalid_login_code

    def test_reminder_cards_info_leak(self):
        """Check that it isn't leaked whether an access code exists."""
        self._test_incorrect_teacher_no_info_leak("teacher_print_reminder_cards")

    def test_class_page_info_leak(self):
        """Check that it isn't leaked whether an access code exists."""
        self._test_incorrect_teacher_no_info_leak("onboarding-class")

    def test_student_edit_info_leak(self):
        c = Client()
        t_email, t_pass = signup_teacher_directly()
        c.login(email=t_email, password=t_pass)
        profile = UserProfile(user=User.objects.create_user("test"))
        profile.save()
        stu = Student(user=profile)
        stu.save()

        assert (
            c.get(reverse("teacher_edit_student", kwargs={"pk": "9999"})).status_code
            == c.get(reverse("teacher_edit_student", kwargs={"pk": stu.pk})).status_code
        )

    def test_cannot_create_school_with_email_as_name(self):
        number_of_existing_schools = len(School.objects.all())

        email, password = signup_teacher_directly()

        client = Client()
        client.login(username=email, password=password)

        url = reverse("onboarding-organisation")
        data = {"name": email, "postcode": "TEST", "country": "GB", "create_organisation": ""}

        client.post(url, data)

        assert number_of_existing_schools == len(School.objects.all())

    def test_reminder_cards_wrong_teacher(self):
        """Try and view reminder cards without being the teacher for that class."""
        self._test_incorrect_teacher_cannot_login("teacher_print_reminder_cards")

    def test_class_page_wrong_teacher(self):
        """Try and view a class page without being the teacher for that class."""
        self._test_incorrect_teacher_cannot_login("onboarding-class")




# Generated by Django 1.11.24 on 2020-09-10 08:46
from __future__ import unicode_literals

from django.db import migrations
from django.db.migrations.operations.special import SeparateDatabaseAndState


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ("portal", "0058_move_to_common_models"),
        ("common", "0002_emailverification"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name="emailverification",
                    name="user",
                ),
            ],
            database_operations=[],
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[migrations.DeleteModel(name="EmailVerification")],
            database_operations=[
                migrations.AlterModelTable(
                    name="EmailVerification", table="common_emailverification"
                )
            ],
        ),
    ]



from django import forms


# This is the form that asks if the user wants to disable 2FA after clicking disable
# 2FA, setting it to always checked and hidden in CSS
class DisableForm(forms.Form):
    understand = forms.BooleanField(label="", initial=True)



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from .base_page import BasePage


class HomePage(BasePage):
    def __init__(self, browser):
        super(HomePage, self).__init__(browser)
        assert self.on_correct_page("home_page")

    def go_to_teacher_login_page(self):
        self.browser.find_element(By.ID, "login_dropdown").click()
        self.browser.find_element(By.ID, "teacher_login_button").click()

        from portal.tests.pageObjects.portal.teacher_login_page import TeacherLoginPage

        return TeacherLoginPage(self.browser)

    def teacher_logout(self):
        self.browser.find_element(By.ID, "logout_menu").click()
        self.browser.find_element(By.ID, "logout_button").click()
        return HomePage(self.browser)

    def go_to_independent_student_login_page(self):
        self.browser.find_element(By.ID, "login_dropdown").click()
        self.browser.find_element(By.ID, "independent_login_button").click()

        from portal.tests.pageObjects.portal.independent_login_page import IndependentStudentLoginPage

        return IndependentStudentLoginPage(self.browser)

    def go_to_student_login_page(self):
        self.browser.find_element(By.ID, "login_dropdown").click()
        self.browser.find_element(By.ID, "student_login_button").click()

        from portal.tests.pageObjects.portal.student_login_class_code import StudentLoginClassCodePage

        return StudentLoginClassCodePage(self.browser)

    def go_to_signup_page(self):
        self.browser.find_element(By.ID, "signup_button").click()

        import portal.tests.pageObjects.portal.signup_page as signup_page

        return signup_page.SignupPage(self.browser)



from selenium.webdriver.common.by import By

from .teach_base_page import TeachBasePage


class TeachClassPage(TeachBasePage):
    def __init__(self, browser):
        super(TeachClassPage, self).__init__(browser)

        assert self.on_correct_page("teach_class_page")

    def type_student_name(self, name):
        self.browser.find_element(By.ID, "id_names").send_keys(name + "\n")
        return self

    def create_students(self):
        self.click_create_students()

        import portal.tests.pageObjects.portal.teach.onboarding_student_list_page as onboarding_student_list_page

        return onboarding_student_list_page.OnboardingStudentListPage(self.browser)

    def click_create_students(self):
        self.browser.find_element(By.NAME, "new_students").click()
        return self

    def adding_students_failed(self):
        if not self.element_exists_by_css(".errorlist"):
            return False

        error_list = self.browser.find_element(By.ID, "form-create-students").find_element(By.CLASS_NAME, "errorlist")

        return error_list.text

    def duplicate_students(self, name):
        if not self.element_exists_by_css(".errorlist"):
            return False

        errors = self.browser.find_element(By.ID, "form-create-students").find_element(By.CLASS_NAME, "errorlist").text
        error = "You cannot add more than one student called '{0}'".format(name)
        return error in errors

    def student_exists(self, name):
        return name in self.browser.find_element(By.ID, "student_table").text

    def delete_students(self):
        self.browser.find_element(By.ID, "deleteSelectedStudents").click()
        return self

    def reset_passwords(self):
        self.browser.find_element(By.ID, "resetSelectedStudents").click()
        return self

    def move_students(self):
        self.browser.find_element(By.ID, "moveSelectedStudents").click()

        import portal.tests.pageObjects.portal.teach.move_students_page as move_students_page

        return move_students_page.TeachMoveStudentsPage(self.browser)

    def move_students_none_selected(self):
        self.browser.find_element(By.ID, "moveSelectedStudents").click()

        return self

    def dismiss_students(self):
        self.browser.find_element(By.ID, "dismissSelectedStudents").click()

        import portal.tests.pageObjects.portal.teach.dismiss_students_page as dismiss_students_page

        return dismiss_students_page.TeachDismissStudentsPage(self.browser)

    def confirm_delete_student_dialog(self):
        self.confirm_dialog()

        return self

    def confirm_reset_student_dialog(self):
        self.confirm_dialog()

        import portal.tests.pageObjects.portal.teach.onboarding_student_list_page as onboarding_student_list_page

        return onboarding_student_list_page.OnboardingStudentListPage(self.browser)

    def toggle_select_student(self):
        self.browser.find_element(By.ID, "student_checkbox").click()
        return self

    def toggle_all_students(self):
        self.browser.find_element(By.ID, "selectedStudentsListToggle").click()
        return self

    def has_students(self):
        return self.element_exists_by_id("student_table")

    def go_to_edit_student_page(self):
        self.browser.find_element(By.ID, "edit_student_button").click()

        import portal.tests.pageObjects.portal.teach.edit_student_page as edit_student_page

        return edit_student_page.EditStudentPage(self.browser)

    def go_to_dashboard(self):
        self.browser.find_element(By.ID, "return_to_classes_button").click()
        # stop scrolling animation and scroll to top as the classes page will scroll to classes by default
        self.browser.execute_script("$('html,body').stop();window.scrollTo(0, 0)")

        import portal.tests.pageObjects.portal.teach.dashboard_page as dashboard_page

        return dashboard_page.TeachDashboardPage(self.browser)



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0001_squashed_0041_new_news")]

    operations = [
        migrations.AddField(
            model_name="school",
            name="country",
            field=models.CharField(max_length=200, null=True, blank=True),
            preserve_default=True,
        )
    ]



from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [("portal", "0048_plural_management_frontnews")]

    operations = [
        migrations.AddField(
            model_name="emailverification",
            name="new_user",
            field=models.ForeignKey(
                to=settings.AUTH_USER_MODEL,
                related_name="email_verifications_old",
                null=True,
                blank=True,
                on_delete=models.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="emailverification",
            name="verified",
            field=models.BooleanField(default=False),
        ),
        migrations.RunSQL(
            "UPDATE portal_emailverification SET new_user_id = ( SELECT portal_userprofile.user_id FROM portal_userprofile WHERE portal_userprofile.id = portal_emailverification.user_id);"
        ),
        migrations.RunSQL("UPDATE portal_emailverification SET verified = used;"),
    ]



import hashlib
from unittest.mock import patch, Mock

from portal.helpers.password import is_password_pwned


class TestClass:
    def test_is_password_pwned(self):
        weak_password = "Password123$"
        strong_password = "£EDCVFR$%TGBnhy667ujm"
        assert is_password_pwned(weak_password)
        assert not is_password_pwned(strong_password)

    @patch("requests.get")
    def test_is_password_pwned__status_code_not_200(self, mock_get):
        # Arrange
        password = "password123"
        sha1_hash = hashlib.sha1(password.encode()).hexdigest()

        mock_response = Mock()
        mock_response.status_code = 500

        mock_get.return_value = mock_response

        # Act
        result = is_password_pwned(password)

        # Assert
        mock_get.assert_called_once_with(f"https://api.pwnedpasswords.com/range/{sha1_hash[:5]}")
        assert not result



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from .teach_base_page import TeachBasePage


class TeachMoveClassesPage(TeachBasePage):
    def __init__(self, browser):
        super(TeachMoveClassesPage, self).__init__(browser)

        assert self.on_correct_page("move_all_classes_page")

    def move_and_kick(self):
        self.browser.find_element(By.ID, "move_classes_button").click()
        import portal.tests.pageObjects.portal.teach.dashboard_page as dashboard_page

        return dashboard_page.TeachDashboardPage(self.browser)

    def move_and_leave(self):
        self.browser.find_element(By.ID, "move_classes_button").click()
        import portal.tests.pageObjects.portal.teach.onboarding_organisation_page as onboarding_organisation_page

        return onboarding_organisation_page.OnboardingOrganisationPage(self.browser)



from django import template

register = template.Library()


@register.inclusion_tag("portal/partials/card_list.html", takes_context=True)
def card_list(context):
    """
    Registers the inclusion tag for the card list partial.
    The template currently expects a list of elements which each contain the following:
    - image: the path to the card's image (top-half)
    - title: the heading of the card
    - description (optional): the text paragraph of the card
    """
    return context["CARD_LIST"]



from __future__ import absolute_import

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select

from . import onboarding_classes_page
from .teach_base_page import TeachBasePage


class OnboardingOrganisationPage(TeachBasePage):
    def __init__(self, browser):
        super(OnboardingOrganisationPage, self).__init__(browser)

        assert self.on_correct_page("onboarding_organisation_page")

    def create_organisation(self, name, password, country="GB"):
        self._create_organisation(name, password, country)

        return onboarding_classes_page.OnboardingClassesPage(self.browser)

    def create_organisation_failure(self, name, password, country="GB"):
        self._create_organisation(name, password, country)

        return self

    def create_organisation_empty(self):
        self._click_create_school_button()

        return self

    def _create_organisation(self, name, password, country):
        self.browser.find_element(By.ID, "id_name").send_keys(name)
        country_element = self.browser.find_element(By.ID, "id_country")
        select = Select(country_element)
        select.select_by_value(country)
        self._click_create_school_button()

    def _click_create_school_button(self):
        self.browser.find_element(By.NAME, "create_organisation").click()

    def has_creation_failed(self):
        if not self.element_exists_by_css(".errorlist"):
            return False

        errors = (
            self.browser.find_element(By.ID, "form-create-organisation").find_element(By.CLASS_NAME, "errorlist").text
        )
        error = "There is already a school or club registered with that name and postcode"
        return error in errors

    def was_postcode_invalid(self):
        errors = (
            self.browser.find_element(By.ID, "form-create-organisation").find_element(By.CLASS_NAME, "errorlist").text
        )
        error = "Please enter a valid postcode or ZIP code"
        return error in errors



import common.permissions as permissions
from django.contrib.auth.decorators import user_passes_test
from django.urls import reverse_lazy
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from two_factor.forms import MethodForm
from two_factor.views.core import SetupView


def login_required(function=None):
    """
    Decorator for views that checks that the user is logged in, redirecting
    to the log-in page if necessary.
    """
    actual_decorator = user_passes_test(
        permissions.logged_in_as_teacher,
        login_url=reverse_lazy("teacher_login"),
    )
    if function:
        return actual_decorator(function)
    return actual_decorator


# This custom class gets rid of the 'welcome' step of 2FA
# which the new design not needs any more
@method_decorator([never_cache, login_required], name="dispatch")
class CustomSetupView(SetupView):
    form_list = (("method", MethodForm),)



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("portal/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="codeforlife-portal",
    version=version,
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django-recaptcha==4.0.0",
        "pyyaml==6.0.2",
        "importlib-metadata==4.13.0",
        "reportlab==4.2.5",
        "django-formtools==2.5.1",
        "django-otp==1.5.4",
        "requests==2.32.2",
        "django-treebeard==4.7.1",
        "django-sekizai==4.1.0",
        "django-classy-tags==4.1.0",
        "phonenumbers==8.12.12",
        "django-ratelimit==3.0.1",
        "django-preventconcurrentlogins==0.8.2",
        "setuptools==74.0.0",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.12",
        "Framework :: Django",
    ],
    zip_safe=False,
)



from __future__ import absolute_import

from datetime import datetime, timedelta

from common.models import Class, DailyActivity, Teacher
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import (
    create_organisation_directly,
    join_teacher_to_organisation,
)
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from django.test import Client, TestCase
from django.urls import reverse
from game.models import Level

from .base_test import BaseTest
from .pageObjects.portal.home_page import HomePage
from .pageObjects.portal.teach.class_page import TeachClassPage
from .utils.classes import create_class
from .utils.messages import is_class_created_message_showing


class TestClass(TestCase):
    def test_delete_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        create_organisation_directly(email1)
        klass, klass_name, access_code = create_class_directly(email1)
        _, _, student = create_school_student_directly(access_code)

        c = Client()

        url = reverse("teacher_delete_class", kwargs={"access_code": access_code})

        # Login as another teacher, try to delete the class and check for 404
        c.login(username=email2, password=password2)

        response = c.post(url)

        assert response.status_code == 404

        c.logout()

        # Login as first teacher, check there is a class
        c.login(username=email1, password=password1)

        teacher = Teacher.objects.get(new_user__username=email1)
        teacher_classes = Class.objects.filter(teacher=teacher)

        assert len(teacher_classes) == 1

        # Try to delete the class, check that it can't be deleted since it's not empty
        response = c.post(url)

        teacher_classes = Class.objects.filter(teacher=teacher)

        assert response.status_code == 302
        assert len(teacher_classes) == 1
        assert teacher.has_class()

        # Delete the student, and try again, check the class is deleted successfully
        student.delete()

        response = c.post(url)

        teacher_classes = Class.objects.filter(teacher=teacher)

        assert response.status_code == 302
        assert len(teacher_classes) == 0
        assert not teacher.has_class()

        # Check class is anonymised
        new_klass = Class._base_manager.get(pk=klass.id)
        assert new_klass.name != klass_name
        assert new_klass.access_code == ""
        assert not new_klass.is_active

    def test_edit_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        create_organisation_directly(email1)
        _, class_name, access_code = create_class_directly(email1)
        create_school_student_directly(access_code)

        c = Client()

        url = reverse("teacher_edit_class", kwargs={"access_code": access_code})
        new_class_name = "New class name"
        data = {
            "name": new_class_name,
            "classmate_progress": "on",
            "external_requests": "1000",  # Setting to always accept requests
            "class_edit_submit": "",
        }

        # Login as another teacher, try to edit the class and check for 404
        c.login(username=email2, password=password2)

        response = c.post(url, data)

        assert response.status_code == 404

        c.logout()

        # Login as first teacher, check the default class settings
        c.login(username=email1, password=password1)

        teacher = Teacher.objects.get(new_user__username=email1)
        klass = Class.objects.get(teacher=teacher)

        assert klass.name == class_name
        assert not klass.classmates_data_viewable
        assert not klass.accept_requests_until

        # Edit class settings, check they match the data dict above
        response = c.post(url, data)

        assert response.status_code == 302

        klass = Class.objects.get(teacher=teacher)

        assert klass.name == new_class_name
        assert klass.classmates_data_viewable
        assert klass.always_accept_requests

    def test_level_control(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass1, _, access_code1 = create_class_directly(email)
        klass2, _, access_code2 = create_class_directly(email)

        c = Client()
        old_date = datetime.now() - timedelta(days=1)
        old_daily_activity = DailyActivity(date=old_date)
        old_daily_activity.save()

        url = reverse("teacher_edit_class", kwargs={"access_code": access_code1})
        # POST request data for locking only the first level
        data = {
            "Getting Started": [
                "level:2",
                "level:3",
                "level:4",
                "level:5",
                "level:6",
                "level:7",
                "level:8",
                "level:9",
                "level:10",
                "level:11",
                "level:12",
            ],
            "Shortest Route": [
                "level:13",
                "level:14",
                "level:15",
                "level:16",
                "level:17",
                "level:18",
            ],
            "Loops and Repetitions": [
                "level:19",
                "level:20",
                "level:21",
                "level:22",
                "level:23",
                "level:24",
                "level:25",
                "level:26",
                "level:27",
                "level:28",
            ],
            "Loops with Conditions": [
                "level:29",
                "level:30",
                "level:31",
                "level:32",
            ],
            "If... Only": [
                "level:33",
                "level:34",
                "level:35",
                "level:36",
                "level:37",
                "level:38",
                "level:39",
                "level:40",
                "level:41",
                "level:42",
                "level:43",
            ],
            "Traffic Lights": [
                "level:44",
                "level:45",
                "level:46",
                "level:47",
                "level:48",
                "level:49",
                "level:50",
            ],
            "Limited Blocks": [
                "level:53",
                "level:78",
                "level:79",
                "level:80",
                "level:81",
                "level:82",
                "level:83",
                "level:84",
                "level:54",
                "level:55",
            ],
            "Procedures": [
                # "level:85",
                # "level:52",
                # "level:60",
                # "level:86",
                # "level:62",
                # "level:87",
                # "level:61",
            ],
            "Blockly Brain Teasers": [
                "level:56",
                "level:57",
                "level:58",
                "level:59",
                "level:88",
                "level:91",
                "level:90",
                "level:89",
                "level:110",
                "level:111",
                "level:112",
                "level:92",
            ],
            "Introduction to Python": [
                "level:93",
                "level:63",
                "level:64",
                "level:65",
                "level:94",
                "level:66",
                "level:67",
                "level:68",
                "level:95",
                "level:69",
                "level:96",
                "level:97",
            ],
            "Python": [
                "level:98",
                "level:70",
                "level:71",
                "level:73",
                "level:72",
                "level:99",
                "level:74",
                "level:75",
                "level:100",
                "level:101",
                "level:102",
                "level:103",
                "level:104",
                "level:105",
                "level:106",
                "level:107",
                "level:108",
                "level:109",
            ],
            "level_control_submit": "",
        }

        c.login(username=email, password=password)

        response = c.post(url, data)

        assert response.status_code == 302

        level1 = Level.objects.get(name=1)

        assert klass1 in level1.locked_for_class.all()
        assert klass2 not in level1.locked_for_class.all()
        messages = list(response.wsgi_request._messages)
        assert len(messages) == 1
        assert str(messages[0]) == "Your level preferences have been saved."

        # test the old analytic stays the same and the new one is incremented
        assert DailyActivity.objects.get(date=old_date).level_control_submits == 0
        assert DailyActivity.objects.get(date=datetime.now()).level_control_submits == 1

        # Resubmitting to unlock level 1
        data = {
            "Getting Started": [
                "level:1",
                "level:2",
                "level:3",
                "level:4",
                "level:5",
                "level:6",
                "level:7",
                "level:8",
                "level:9",
                "level:10",
                "level:11",
                "level:12",
            ],
            "Shortest Route": [
                "level:13",
                "level:14",
                "level:15",
                "level:16",
                "level:17",
                "level:18",
            ],
            "Loops and Repetitions": [
                "level:19",
                "level:20",
                "level:21",
                "level:22",
                "level:23",
                "level:24",
                "level:25",
                "level:26",
                "level:27",
                "level:28",
            ],
            "Loops with Conditions": [
                "level:29",
                "level:30",
                "level:31",
                "level:32",
            ],
            "If... Only": [
                "level:33",
                "level:34",
                "level:35",
                "level:36",
                "level:37",
                "level:38",
                "level:39",
                "level:40",
                "level:41",
                "level:42",
                "level:43",
            ],
            "Traffic Lights": [
                "level:44",
                "level:45",
                "level:46",
                "level:47",
                "level:48",
                "level:49",
                "level:50",
            ],
            "Limited Blocks": [
                "level:53",
                "level:78",
                "level:79",
                "level:80",
                "level:81",
                "level:82",
                "level:83",
                "level:84",
                "level:54",
                "level:55",
            ],
            "Procedures": [
                # "level:85",
                # "level:52",
                # "level:60",
                # "level:86",
                # "level:62",
                # "level:87",
                # "level:61",
            ],
            "Blockly Brain Teasers": [
                "level:56",
                "level:57",
                "level:58",
                "level:59",
                "level:88",
                "level:91",
                "level:90",
                "level:89",
                "level:110",
                "level:111",
                "level:112",
                "level:92",
            ],
            "Introduction to Python": [
                "level:93",
                "level:63",
                "level:64",
                "level:65",
                "level:94",
                "level:66",
                "level:67",
                "level:68",
                "level:95",
                "level:69",
                "level:96",
                "level:97",
            ],
            "Python": [
                "level:98",
                "level:70",
                "level:71",
                "level:73",
                "level:72",
                "level:99",
                "level:74",
                "level:75",
                "level:100",
                "level:101",
                "level:102",
                "level:103",
                "level:104",
                "level:105",
                "level:106",
                "level:107",
                "level:108",
                "level:109",
            ],
            "level_control_submit": "",
        }

        response = c.post(url, data)

        assert response.status_code == 302

        level1 = Level.objects.get(name=1)

        assert klass1 not in level1.locked_for_class.all()
        assert klass2 not in level1.locked_for_class.all()

    def test_transfer_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        school = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, school.name)
        klass1, _, access_code1 = create_class_directly(email1)
        klass2, _, access_code2 = create_class_directly(email2)
        _, _, student1 = create_school_student_directly(access_code1)
        _, _, student2 = create_school_student_directly(access_code2)

        teacher1 = Teacher.objects.get(new_user__username=email1)
        teacher2 = Teacher.objects.get(new_user__username=email2)
        teacher1_classes = Class.objects.filter(teacher=teacher1)
        teacher2_classes = Class.objects.filter(teacher=teacher2)

        assert len(teacher1_classes) == 1
        assert len(teacher2_classes) == 1

        c = Client()

        url = reverse("teacher_edit_class", kwargs={"access_code": access_code1})
        data = {"new_teacher": teacher2.id, "class_move_submit": ""}

        # Login as first teacher and transfer class to the second teacher
        c.login(username=email1, password=password1)

        response = c.post(url, data)

        assert response.status_code == 302

        c.logout()

        teacher1_classes = Class.objects.filter(teacher=teacher1)
        teacher2_classes = Class.objects.filter(teacher=teacher2)

        assert len(teacher1_classes) == 0
        assert len(teacher2_classes) == 2
        assert teacher2_classes[1] == klass2
        assert not teacher1.teaches(student2.user)
        assert teacher2.teaches(student2.user)


# Class for Selenium tests. We plan to replace these and turn them into Cypress tests
class TestClassFrontend(BaseTest):
    def test_create(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        page = self.go_to_homepage().go_to_teacher_login_page().login_no_class(email, password).open_classes_tab()

        assert page.does_not_have_classes()

        page, class_name = create_class(page)
        assert is_class_created_message_showing(self.selenium, class_name)

    def test_create_class_as_admin_for_another_teacher(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        teacher2 = Teacher.objects.get(new_user__email=email2)
        school = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, school.name)

        # Check teacher 2 doesn't have any classes
        page = self.go_to_homepage().go_to_teacher_login_page().login(email2, password2).open_classes_tab()
        assert page.does_not_have_classes()
        page.logout()

        # Log in as the first teacher and create a class for the second one
        page = self.go_to_homepage().go_to_teacher_login_page().login(email1, password1).open_classes_tab()
        page, class_name = create_class(page, teacher_id=teacher2.id)
        page = TeachClassPage(page.browser)
        assert is_class_created_message_showing(self.selenium, class_name)
        page.logout()

        # Check teacher 2 now has the class
        page = self.go_to_homepage().go_to_teacher_login_page().login(email2, password2).open_classes_tab()
        assert page.has_classes()

    def test_create_dashboard(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        page = self.go_to_homepage().go_to_teacher_login_page().login(email, password).open_classes_tab()

        page, class_name = create_class(page)

        assert is_class_created_message_showing(self.selenium, class_name)

    def test_create_dashboard_non_admin(self):
        email_1, password_1 = signup_teacher_directly()
        email_2, password_2 = signup_teacher_directly()
        school = create_organisation_directly(email_1)
        klass_1, class_name_1, access_code_1 = create_class_directly(email_1)
        create_school_student_directly(access_code_1)
        join_teacher_to_organisation(email_2, school.name)
        klass_2, class_name_2, access_code_2 = create_class_directly(email_2)
        create_school_student_directly(access_code_2)

        page = self.go_to_homepage().go_to_teacher_login_page().login(email_2, password_2).open_classes_tab()

        page, class_name_3 = create_class(page)

        assert is_class_created_message_showing(self.selenium, class_name_3)

    def test_create_invalid_name(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)

        class_name = "Class!"

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_class(email, password)
            .open_classes_tab()
        )

        page = page.create_class(class_name, False)

        assert page.was_form_invalid(
            "form-create-class", "Class name may only contain letters, numbers, dashes, underscores, and spaces."
        )



from __future__ import absolute_import

from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly

from portal.tests.pageObjects.portal.home_page import HomePage
from .base_test import BaseTest
from .utils.messages import is_student_details_updated_message_showing, is_password_updated_message_showing


class TestSchoolStudent(BaseTest):
    def test_login(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

    def test_login_invalid_class_code(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_student_login_page().student_input_access_code_failure("not a class code")

        assert page.has_access_code_input_failed(
            "form-login-school-class-code", "Uh oh! You didn't input a valid class code."
        )

    def test_login_failure(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login_failure(student_name, "some other password")
        )

        assert page.has_login_failed("form-login-school", "Invalid name, class access code or password")

    def test_login_nonexistent_class(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code("WRONG")
            .student_login_failure(student_name, student_password)
        )

        assert page.has_login_failed("form-login-school", "Invalid name, class access code or password")

    def test_login_empty_class(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)
        _, _, access_code2 = create_class_directly(email)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code2)
            .student_login_failure(student_name, student_password)
        )

        assert page.has_login_failed("form-login-school", "Invalid name, class access code or password")

    def test_update_password_current_password_wrong(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

        page = page.go_to_account_page().update_password_failure(
            "£EDCVFR$5tgb", "£EDCVFR$5tgb", "Wrong_123$£$3_Password"
        )
        assert self.is_account_page(page)
        assert page.was_form_invalid("student_account_form", "Your current password was incorrect")

    def test_update_password_passwords_not_match(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

        page = page.go_to_account_page().update_password_failure("£EDECVFR$5tgb", "%TGBNHY^&ujm,ki8", student_password)
        assert self.is_account_page(page)
        assert page.was_form_invalid("student_account_form", "Your new passwords do not match")

    def test_update_password_too_weak(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

        page = page.go_to_account_page().update_password_failure("tiny", "tiny", student_password)
        assert self.is_account_page(page)
        assert page.was_form_invalid(
            "student_account_form",
            "Password not strong enough, consider using at least 6 characters and making it hard to guess.",
        )

    def test_update_password_too_common(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

        page = page.go_to_account_page().update_password_failure("Password123$", "Password123$", student_password)
        assert self.is_account_page(page)
        assert page.was_form_invalid(
            "student_account_form", "Password is too common, consider using a different password."
        )

    def test_update_password_success(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_student_login_page()
            .student_input_access_code(access_code)
            .student_login(student_name, student_password)
        )
        assert self.is_dashboard(page)

        new_password = "£EDCVFR$%TGBhny6"

        page = page.go_to_account_page().update_password_success(new_password, student_password)
        assert is_student_details_updated_message_showing(self.selenium)
        assert is_password_updated_message_showing(self.selenium)
        assert self.is_login_class_code_page(page)

        page = page.student_input_access_code(access_code).student_login(student_name, new_password)
        assert self.is_dashboard(page)

    def is_dashboard(self, page):
        return page.__class__.__name__ == "PlayDashboardPage"

    def is_account_page(self, page):
        return page.__class__.__name__ == "PlayAccountPage"

    def is_login_class_code_page(self, page):
        return page.__class__.__name__ == "StudentLoginClassCodePage"



import pytest


@pytest.mark.django_db
def test_verify_portaladmin(migrator):
    migrator.apply_initial_migration(("portal", "0061_make_portaladmin_teacher"))
    new_state = migrator.apply_tested_migration(("portal", "0062_verify_portaladmin"))

    User = new_state.apps.get_model("auth", "User")
    EmailVerification = new_state.apps.get_model("common", "EmailVerification")

    portaladmin = User.objects.get(username="portaladmin")
    portaladmin_verification = EmailVerification.objects.get(user=portaladmin)

    assert portaladmin_verification.verified


@pytest.mark.django_db
def test_verify_portaladmin_rollback(migrator):
    migrator.apply_initial_migration(("portal", "0062_verify_portaladmin"))
    new_state = migrator.apply_tested_migration(("portal", "0061_make_portaladmin_teacher"))

    User = new_state.apps.get_model("auth", "User")
    EmailVerification = new_state.apps.get_model("common", "EmailVerification")

    portaladmin = User.objects.get(username="portaladmin")
    assert not EmailVerification.objects.filter(user=portaladmin).exists()



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_banner 1'] = '''<div class="banner banner--teacher">
    <div class="container">
        <div class="row">
            <div class="col-sm-12 d-flex">
                <div>
                    <h1 class="banner__text--primary">Test title</h1>
                    
                        <h4>Test subtitle</h4>
                    
                    
                        <p>Test text</p>
                    
                    <div class="button-group">
                        
                        
                    </div>
                </div>
                <div>
                    <div class="banner--picture">
                        <div class="banner--picture__inside1">
                            <div class="banner--picture__inside2 test--image--class"
                                 
                                 >
                            </div>
                        </div>
                    </div>
                </div>
            </div>
        </div>
    </div>
</div>
'''

snapshots['test_benefits 1'] = '''

<div class="grid-benefits col-sm-8 col-center">
    
    
    
    
        <h5 class="grid-benefits__title grid-benefits__title1">Test title</h5>
    
    
        <h5 class="grid-benefits__title grid-benefits__title2">Test title</h5>
    
    
        <h5 class="grid-benefits__title grid-benefits__title3">Test title</h5>
    
    <p class="grid-benefits__text1">Test text</p>
    <p class="grid-benefits__text2">Test text</p>
    <p class="grid-benefits__text3">Test text</p>
    
        <div class="grid-benefits__button grid-benefits__button1">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
    
        <div class="grid-benefits__button grid-benefits__button2">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
    
        <div class="grid-benefits__button grid-benefits__button3">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
</div>
'''

snapshots['test_headline 1'] = '''<section>
    <h4>Test title</h4>
</section>
<p class="container">Test description</p>
'''



import csv
import io
import json
from datetime import date, datetime, timedelta
from unittest.mock import ANY, Mock, patch

import pytest
from common.mail import campaign_ids
from common.models import (
    Class,
    DailyActivity,
    School,
    Student,
    Teacher,
    TotalActivity,
    UserProfile,
    UserSession,
)
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import (
    create_organisation_directly,
    join_teacher_to_organisation,
)
from common.tests.utils.student import (
    create_independent_student_directly,
    create_school_student_directly,
    create_student_with_direct_login,
)
from common.tests.utils.teacher import signup_teacher_directly
from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse
from django.utils import timezone
from game.models import Level
from game.tests.utils.attempt import create_attempt
from game.tests.utils.level import create_save_level
from pypdf import PdfReader
from rest_framework.test import APIClient, APITestCase

from deploy import captcha
from portal.templatetags.app_tags import is_logged_in_as_admin_teacher
from portal.views.api import anonymise
from portal.views.cron.user import (
    USER_DELETE_UNVERIFIED_ACCOUNT_DAYS,
    USER_2ND_INACTIVE_REMINDER_DAYS,
    USER_FINAL_INACTIVE_REMINDER_DAYS,
    USER_RETENTION_PERIOD,
)
from portal.views.teacher.teach import (
    REMINDER_CARDS_PDF_COLUMNS,
    REMINDER_CARDS_PDF_ROWS,
    REMINDER_CARDS_PDF_WARNING_TEXT,
    count_student_details_click,
)


class TestTeacherViews(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.email, cls.password = signup_teacher_directly()
        cls.school = create_organisation_directly(cls.email)
        _, _, cls.class_access_code = create_class_directly(cls.email)
        _, cls.password_student, cls.student = create_school_student_directly(
            cls.class_access_code
        )

    def login(self):
        c = Client()
        assert c.login(username=self.email, password=self.password)
        return c

    def test_reminder_cards(self):
        c = self.login()
        url = reverse(
            "teacher_print_reminder_cards", args=[self.class_access_code]
        )

        # First test with 2 dummy students
        NAME1 = "Test name"
        NAME2 = "Another name"
        PASSWORD1 = "password1"
        PASSWORD2 = "password2"
        URL = "url"

        studentlist = [
            {"name": NAME1, "password": PASSWORD1, "login_url": URL},
            {"name": NAME2, "password": PASSWORD2, "login_url": URL},
        ]
        data = {"data": json.dumps(studentlist)}

        response = c.post(url, data)
        assert response.status_code == 200

        # read PDF, check there's only 1 page and that the correct student details show
        with io.BytesIO(response.content) as pdf_file:
            file_reader = PdfReader(pdf_file)
            assert len(file_reader.pages) == 1

            page_text = file_reader.pages[0].extract_text()
            assert NAME1 in page_text
            assert NAME2 in page_text
            assert PASSWORD1 in page_text
            assert PASSWORD2 in page_text
            assert REMINDER_CARDS_PDF_WARNING_TEXT in page_text

        # Add students to the dummy data list until it goes over the max students per
        # page number
        students_per_page = REMINDER_CARDS_PDF_ROWS * REMINDER_CARDS_PDF_COLUMNS
        for _ in range(len(studentlist), students_per_page + 1):
            studentlist.append(
                {"name": NAME1, "password": PASSWORD1, "login_url": URL}
            )

        assert len(studentlist) == students_per_page + 1

        data = {"data": json.dumps(studentlist)}
        response = c.post(url, data)
        assert response.status_code == 200

        # Check there are 2 pages and that each page contains the warning text
        with io.BytesIO(response.content) as pdf_file:
            file_reader = PdfReader(pdf_file)
            assert len(file_reader.pages) == 2

            page1_text = file_reader.pages[0].extract_text()
            page2_text = file_reader.pages[1].extract_text()
            assert REMINDER_CARDS_PDF_WARNING_TEXT in page1_text
            assert REMINDER_CARDS_PDF_WARNING_TEXT in page2_text

    def test_csv(self):
        c = self.login()
        url = reverse("teacher_download_csv", args=[self.class_access_code])
        NAME1 = "Test name"
        NAME2 = "Another name"
        PASSWORD = "easy"
        URL_PLACEHOLDER = "http://_____"

        studentlist = [
            {"name": NAME1, "password": PASSWORD, "login_url": URL_PLACEHOLDER},
            {"name": NAME2, "password": PASSWORD, "login_url": URL_PLACEHOLDER},
        ]
        data = {"data": json.dumps(studentlist)}

        response = c.post(url, data)

        assert response.status_code == 200
        content = response.content.decode("utf-8")
        reader = csv.reader(io.StringIO(content))

        access_code = self.class_access_code
        class_url = reverse(
            "student_login", kwargs={"access_code": access_code}
        )
        row0 = next(reader)
        assert row0[0].strip() == access_code
        assert class_url in row0[1].strip()
        row1 = next(reader)
        assert row1[0] == NAME1
        assert row1[1] == PASSWORD
        assert row1[2] == URL_PLACEHOLDER
        row2 = next(reader)
        assert row2[0] == NAME2

        # post without any data should return empty
        response = c.post(url)
        assert response.status_code == 200
        content = response.content.decode("utf-8")
        assert content == ""

        # as well as GET method
        response = c.get(url)
        assert response.status_code == 200
        content = response.content.decode("utf-8")
        assert content == ""

    def test_organisation_kick_has_correct_permissions(self):
        teacher2_email, _ = signup_teacher_directly()
        school = create_organisation_directly(self.email)
        join_teacher_to_organisation(self.email, school.name, is_admin=True)
        join_teacher_to_organisation(teacher2_email, school.name)
        teacher2_id = Teacher.objects.get(new_user__email=teacher2_email).id

        client = self.login()
        url = reverse("organisation_kick", args=[teacher2_id])
        response = client.get(url)
        assert response.status_code == 405
        response = client.post(url)
        assert response.status_code == 302

    def test_daily_activity_student_details(self):
        c = self.login()
        url = reverse(
            "teacher_print_reminder_cards", args=[self.class_access_code]
        )

        data = {
            "data": json.dumps(
                [
                    {
                        "name": self.student.new_user.first_name,
                        "password": self.student.new_user.password,
                        "login_url": self.student.login_id,
                    }
                ]
            )
        }

        # Check there are no DailyActivity rows
        assert DailyActivity.objects.count() == 0

        # Load student login cards
        response = c.post(url, data)
        assert response.status_code == 200

        # Expect a DailyActivity to have been created for today, and expect login cards
        # count to have been incremented
        assert DailyActivity.objects.count() == 1
        daily_activity = DailyActivity.objects.get(id=1)
        assert daily_activity.date == date.today()
        assert daily_activity.csv_click_count == 0
        assert daily_activity.login_cards_click_count == 1

        url = reverse("teacher_download_csv", args=[self.class_access_code])

        # Download student details CSV
        response = c.post(url, data)
        assert response.status_code == 200

        # Expect the same DailyActivity row to be there (no new rows), expect CSV click
        # count to have been incremented and login cards to stay the same
        assert DailyActivity.objects.count() == 1
        daily_activity = DailyActivity.objects.get(id=1)
        assert daily_activity.date == date.today()
        assert daily_activity.csv_click_count == 1
        assert daily_activity.login_cards_click_count == 1

        with pytest.raises(Exception):
            count_student_details_click("Wrong download method")

    def test_release_verified_student(self):
        c = Client()
        student_login_url = reverse(
            "student_login", args=[self.class_access_code]
        )
        response = c.post(
            student_login_url,
            {
                "username": self.student.new_user.first_name,
                "password": self.password_student,
            },
        )
        assert response.status_code == 302

        student = Student.objects.get(pk=self.student.pk)

        assert student.user.is_verified

        c.logout()
        c.login(username=self.email, password=self.password)

        teacher = Teacher.objects.factory(
            "the", "teacher", "theteacher@foo.com", "password"
        )
        level = Level.objects.create()

        level.owner = student.new_user.userprofile
        level.shared_with.add(teacher.new_user)
        level.save()

        students_levels = Level.objects.filter(
            owner=student.new_user.userprofile
        ).all()

        for level in students_levels.all():
            assert level.shared_with.exists()

        release_url = reverse(
            "teacher_dismiss_students", args=[self.class_access_code]
        )
        response = c.post(
            release_url,
            {
                "form-TOTAL_FORMS": 1,
                "form-INITIAL_FORMS": 1,
                "form-MIN_NUM_FORMS": 0,
                "form-MAX_NUM_FORMS": 1000,
                "form-0-orig_name": self.student.new_user.first_name,
                "form-0-name": self.student.new_user.first_name,
                "form-0-email": "independent@gmail.com",
                "form-0-confirm_email": "independent@gmail.com",
                "submit_dismiss": "",
            },
        )
        assert response.status_code == 302

        student = Student.objects.get(pk=self.student.pk)
        assert not student.user.is_verified

        students_levels = Level.objects.filter(
            owner=student.new_user.userprofile
        ).all()

        for level in students_levels.all():
            assert not level.shared_with.exists()


class TestLoginViews(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(TestLoginViews, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(TestLoginViews, cls).tearDownClass()

    def _set_up_test_data(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        _, _, class_access_code = create_class_directly(teacher_email)
        student_name, student_password, _ = create_school_student_directly(
            class_access_code
        )

        return (
            teacher_email,
            teacher_password,
            student_name,
            student_password,
            class_access_code,
        )

    def _create_and_login_teacher(self, next_url=False):
        email, password, _, _, _ = self._set_up_test_data()

        if next_url:
            url = reverse("teacher_login") + "?next=/"
        else:
            url = reverse("teacher_login")

        c = Client()
        response = c.post(
            url,
            {
                "auth-username": email,
                "auth-password": password,
                "teacher_login_view-current_step": "auth",
            },
        )
        return response, c

    def _create_and_login_school_student(self, next_url=False):
        _, _, name, password, class_access_code = self._set_up_test_data()

        if next_url:
            url = (
                reverse(
                    "student_login", kwargs={"access_code": class_access_code}
                )
                + "?next=/"
            )
        else:
            url = reverse(
                "student_login", kwargs={"access_code": class_access_code}
            )

        c = Client()
        response = c.post(url, {"username": name, "password": password})
        return response, c

    def test_teacher_login_redirect(self):
        response, _ = self._create_and_login_teacher(True)
        self.assertRedirects(response, "/")

    def test_student_login_redirect(self):
        response, _ = self._create_and_login_school_student(True)
        self.assertRedirects(response, "/")

    def test_teacher_session(self):
        email, password, _, _, _ = self._set_up_test_data()
        c = Client()
        c.post(
            reverse("teacher_login"),
            {
                "auth-username": email,
                "auth-password": password,
                "teacher_login_view-current_step": "auth",
            },
        )
        # check if there's a UserSession data within the last minute
        now = timezone.now()
        oneminago = now - timedelta(minutes=1)

        user = User.objects.get(email=email)
        q = UserSession.objects.filter(user=user)
        q = q.filter(login_time__range=(oneminago, now))
        assert len(q) == 1

        teacher = Teacher.objects.get(new_user=user)
        assert q[0].school == teacher.school

    def _get_user_class(self, name, class_access_code):
        klass = Class.objects.get(access_code=class_access_code)
        students = Student.objects.filter(
            new_user__first_name__iexact=name, class_field=klass
        )
        assert len(students) == 1
        user = students[0].new_user
        return user, klass

    def test_student_session_class_form(self):
        """Login via class form"""
        _, _, name, password, class_access_code = self._set_up_test_data()
        c = Client()

        resp = c.post(
            reverse("student_login_access_code"),
            {"access_code": class_access_code},
        )
        assert resp.status_code == 302
        nexturl = resp.url
        assert nexturl == reverse(
            "student_login",
            kwargs={
                "access_code": class_access_code,
                "login_type": "classform",
            },
        )
        c.post(nexturl, {"username": name, "password": password})

        # check if there's a UserSession data within the last 10 secs
        now = timezone.now()
        markedtime = now - timedelta(seconds=10)

        user, klass = self._get_user_class(name, class_access_code)

        q = UserSession.objects.filter(user=user)
        q = q.filter(login_time__range=(markedtime, now))
        assert len(q) == 1
        assert q[0].class_field == klass
        assert q[0].login_type == "classform"

    def test_student_session_class_link(self):
        """Login via class link"""
        _, _, name, password, class_access_code = self._set_up_test_data()

        c = Client()
        url = reverse(
            "student_login", kwargs={"access_code": class_access_code}
        )
        c.post(url, {"username": name, "password": password})

        # check if there's a UserSession data within the last 10 secs
        now = timezone.now()
        markedtime = now - timedelta(seconds=10)

        user, klass = self._get_user_class(name, class_access_code)

        q = UserSession.objects.filter(user=user)
        q = q.filter(login_time__range=(markedtime, now))
        assert len(q) == 1
        assert q[0].class_field == klass
        assert q[0].login_type == "classlink"

    def test_student_login_failed(self):
        """Failed login via class link"""
        _, _, name, password, class_access_code = self._set_up_test_data()
        randomname = "randomname"

        c = Client()
        url = reverse(
            "student_login", kwargs={"access_code": class_access_code}
        )
        c.post(url, {"username": randomname, "password": "xx"})

        # check if there's a UserSession data within the last 10 secs
        now = timezone.now()
        markedtime = now - timedelta(seconds=10)

        q = UserSession.objects.filter(login_time__range=(markedtime, now))
        assert len(q) == 0  # login data not found

    def test_indep_student_session(self):
        username, password, student = create_independent_student_directly()
        c = Client()
        url = reverse("independent_student_login")
        c.post(url, {"username": username, "password": password})
        # check if there's a UserSession data within the last minute
        now = timezone.now()
        oneminago = now - timedelta(minutes=1)

        user = User.objects.get(username=username)
        q = UserSession.objects.filter(user=user)
        q = q.filter(login_time__range=(oneminago, now))
        assert len(q) == 1

    def test_student_direct_login(self):
        _, _, _, _, class_access_code = self._set_up_test_data()
        student, login_id, _, _ = create_student_with_direct_login(
            class_access_code
        )

        c = Client()
        assert c.login(user_id=student.new_user.id, login_id=login_id) == True

        url = f"/u/{student.user.id}/{login_id}/"
        response = c.get(url)
        # assert redirects
        assert response.url == "/play/details/"
        assert response.status_code == 302

        # incorrect url
        url = "/u/123/4567890/"
        response = c.get(url)
        assert response.url == "/"
        assert response.status_code == 302

        # check if there's a UserSession data within the last minute
        now = timezone.now()
        oneminago = now - timedelta(minutes=1)
        q = UserSession.objects.filter(user=student.new_user)
        q = q.filter(login_time__range=(oneminago, now))
        assert len(q) == 1
        klass = Class.objects.get(access_code=class_access_code)
        assert q[0].class_field == klass
        assert q[0].login_type == "direct"

    def test_teacher_already_logged_in_register_page_redirect(self):
        _, c = self._create_and_login_teacher()

        url = reverse("register")
        response = c.get(url)
        self.assertRedirects(response, "/teach/dashboard/")

    def test_student_already_logged_in_register_page_redirect(self):
        _, c = self._create_and_login_school_student()

        url = reverse("register")
        response = c.get(url)
        self.assertRedirects(response, "/play/details/")


class TestViews(TestCase):
    def test_home_learning(self):
        c = Client()
        home_url = reverse("home")
        response = c.get(home_url)

        bytes = response.__dict__["_container"][0]
        html = bytes.decode("utf8")

        page_url = reverse("home-learning")

        expected_html = '<a href="/home-learning">Home learning</a>'

        assert expected_html in html

        response = c.get(page_url)

        assert response.status_code == 200

    def test_contributor(self):
        c = Client()
        page_url = reverse("getinvolved")
        response = c.get(page_url)
        assert response.status_code == 200

        page_url = reverse("contribute")
        response = c.get(page_url)
        assert response.status_code == 200

    def test_ten_year_map(self):
        c = Client()
        page_url = reverse("celebrate")
        response = c.get(page_url)
        assert response.status_code == 200

    def test_student_dashboard_view(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        klass, _, class_access_code = create_class_directly(teacher_email)
        (
            student_name,
            student_password,
            student,
        ) = create_school_student_directly(class_access_code)

        # Expected context data when a student hasn't played anything yet
        EXPECTED_DATA_FIRST_LOGIN = {
            "rapid_router": {
                "num_completed": 0,
                "num_top_scores": 0,
                "total_score": 0,
                "total_available_score": 1450,
            },
            "python_den": {
                "num_completed": 0,
                "num_top_scores": 0,
                "total_score": 0,
                "total_available_score": 830,
            },
        }

        # Expected context data when a student has attempted some RR levels
        EXPECTED_DATA_WITH_ATTEMPTS = {
            "rapid_router": {
                "num_completed": 2,
                "num_top_scores": 1,
                "total_score": 39,
                "total_available_score": 1450,
            },
            "python_den": {
                "num_completed": 2,
                "num_top_scores": 2,
                "total_score": 30,
                "total_available_score": 830,
            },
        }

        # Expected context data when a student has also attempted some custom RR
        # levels
        EXPECTED_DATA_WITH_CUSTOM_ATTEMPTS = {
            "rapid_router": {
                "num_completed": 2,
                "num_top_scores": 1,
                "total_score": 39,
                "total_available_score": 1450,
                "total_custom_score": 10,
                "total_custom_available_score": 20,
            },
            "python_den": {
                "num_completed": 2,
                "num_top_scores": 2,
                "total_score": 30,
                "total_available_score": 830,
            },
        }

        c = Client()

        # Login and check initial data
        url = reverse(
            "student_login", kwargs={"access_code": class_access_code}
        )
        c.post(url, {"username": student_name, "password": student_password})

        student_dashboard_url = reverse("student_details")
        response = c.get(student_dashboard_url)

        assert response.status_code == 200
        assert response.context_data == EXPECTED_DATA_FIRST_LOGIN

        # Attempt the first two RR levels, one perfect attempt, one not
        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")

        create_attempt(student, level1, 20)
        create_attempt(student, level2, 19)

        # Attempt the first and fourth Python Den levels, both perfect
        level1001 = Level.objects.get(name="1001")
        level1004 = Level.objects.get(name="1004")

        create_attempt(student, level1001, 20)
        create_attempt(student, level1004, 10)

        response = c.get(student_dashboard_url)

        assert response.status_code == 200
        assert response.context_data == EXPECTED_DATA_WITH_ATTEMPTS

        # Teacher creates 3 custom levels, only shares the first 2 with the
        # student. Check that the total available score only includes the
        # levels shared with the student. Student attempts one level only.
        custom_level1_id = create_save_level(student.class_field.teacher)
        custom_level2_id = create_save_level(student.class_field.teacher)
        create_save_level(student.class_field.teacher)
        custom_level1 = Level.objects.get(id=custom_level1_id.id)
        custom_level2 = Level.objects.get(id=custom_level2_id.id)

        student.new_user.shared.add(custom_level1, custom_level2)
        student.new_user.save()

        create_attempt(student, custom_level2, 10)

        response = c.get(student_dashboard_url)

        assert response.status_code == 200
        assert response.context_data == EXPECTED_DATA_WITH_CUSTOM_ATTEMPTS

    @patch("portal.views.registration.send_dotdigital_email")
    def test_delete_account(self, mock_send_dotdigital_email: Mock):
        email, password = signup_teacher_directly()
        u = User.objects.get(email=email)
        usrid = u.id

        c = Client()
        url = reverse("teacher_login")
        c.post(
            url,
            {
                "auth-username": email,
                "auth-password": password,
                "teacher_login_view-current_step": "auth",
            },
        )

        # fail to delete with incorrect password
        url = reverse("delete_account")
        response = c.post(url, {"password": "wrongPassword"})

        assert response.status_code == 302
        mock_send_dotdigital_email.assert_not_called()

        # user has not been anonymised
        u = User.objects.get(email=email)
        assert u.id == usrid

        # try again with the correct password
        url = reverse("delete_account")
        response = c.post(
            url, {"password": password, "unsubscribe_newsletter": "on"}
        )

        assert response.status_code == 302
        mock_send_dotdigital_email.assert_called_once()
        assert response.url == reverse("home")

        # user has been anonymised
        u = User.objects.get(id=usrid)
        assert u.first_name == "Deleted"
        assert not u.is_active

        assert c.login(username=email, password=password) == False

    @patch("portal.views.registration.send_dotdigital_email")
    def test_delete_account_admin(self, mock_send_dotdigital_email: Mock):
        """test the passing of admin role after deletion of an admin account"""

        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        email3, password3 = signup_teacher_directly()
        email4, password4 = signup_teacher_directly()

        user1 = User.objects.get(email=email1)
        user1.last_name = "Amir"
        user1.save()
        usrid1 = user1.id

        user2 = User.objects.get(email=email2)
        user2.last_name = "Bee"
        user2.save()
        usrid2 = user2.id

        user3 = User.objects.get(email=email3)
        user3.last_name = "Jung"
        user3.save()
        usrid3 = user3.id

        user4 = User.objects.get(email=email4)
        user4.last_name = "Kook"
        user4.save()
        usrid4 = user4.id

        school = create_organisation_directly(email1)
        klass, class_name, access_code_1 = create_class_directly(email1)
        class_id = klass.id
        _, _, student = create_school_student_directly(access_code_1)
        student_user_id = student.new_user.id

        join_teacher_to_organisation(email2, school.name)
        _, _, access_code_2 = create_class_directly(email2)
        create_school_student_directly(access_code_2)

        join_teacher_to_organisation(email3, school.name)
        join_teacher_to_organisation(email4, school.name)

        c = Client()
        url = reverse("teacher_login")
        c.post(
            url,
            {
                "auth-username": email1,
                "auth-password": password1,
                "teacher_login_view-current_step": "auth",
            },
        )

        # delete teacher1 account
        url = reverse("delete_account")
        c.post(url, {"password": password1})
        mock_send_dotdigital_email.assert_called_once()

        # user has been anonymised
        u = User.objects.get(id=usrid1)
        assert not u.is_active

        # check that the class and student have been anonymised
        assert not Class._base_manager.get(pk=class_id).is_active
        student_user1 = User.objects.get(id=student_user_id)
        assert not student_user1.is_active

        school_id = school.id
        school_name = school.name
        teachers = Teacher.objects.filter(school=school).order_by(
            "new_user__last_name", "new_user__first_name"
        )
        assert len(teachers) == 3

        # one of the remaining teachers should be admin (the second in our case, as it's alphabetical)
        u = User.objects.get(id=usrid2)
        assert u.new_teacher.is_admin
        u = User.objects.get(id=usrid3)
        assert not u.new_teacher.is_admin
        u = User.objects.get(id=usrid4)
        assert not u.new_teacher.is_admin

        # make teacher 3 admin
        user3.new_teacher.is_admin = True
        user3.new_teacher.save()

        url = reverse("teacher_login")
        c.post(
            url,
            {
                "auth-username": email3,
                "auth-password": password3,
                "teacher_login_view-current_step": "auth",
            },
        )

        # now delete teacher3 account
        url = reverse("delete_account")
        c.post(url, {"password": password3})
        self.assertEqual(mock_send_dotdigital_email.call_count, 2)

        # 2 teachers left
        teachers = Teacher.objects.filter(school=school).order_by(
            "new_user__last_name", "new_user__first_name"
        )
        assert len(teachers) == 2

        # teacher2 should still be admin, teacher4 is not passed admin role because there is teacher2
        u = User.objects.get(id=usrid2)
        assert u.new_teacher.is_admin
        u = User.objects.get(id=usrid4)
        assert not u.new_teacher.is_admin

        # delete teacher4
        anonymise(user4)

        teachers = Teacher.objects.filter(school=school).order_by(
            "new_user__last_name", "new_user__first_name"
        )
        assert len(teachers) == 1
        u = User.objects.get(id=usrid2)
        assert u.new_teacher.is_admin

        # delete teacher2 (the last one left)
        url = reverse("teacher_login")
        c.post(
            url,
            {
                "auth-username": email2,
                "auth-password": password2,
                "teacher_login_view-current_step": "auth",
            },
        )

        url = reverse("delete_account")
        c.post(url, {"password": password2})
        self.assertEqual(mock_send_dotdigital_email.call_count, 3)

        # school should be anonymised
        school = School._base_manager.get(id=school_id)
        assert school.name != school_name
        assert not school.is_active

        with pytest.raises(School.DoesNotExist):
            School.objects.get(id=school_id)

    def test_legal_pages_load(self):
        c = Client()

        assert c.get(reverse("privacy_notice")).status_code == 200
        assert c.get(reverse("terms")).status_code == 200

    def test_logged_in_as_admin_check(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        school = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, school.name)

        teacher1 = Teacher.objects.get(new_user__username=email1)
        teacher2 = Teacher.objects.get(new_user__username=email2)

        c = Client()

        c.login(username=email1, password=password1)

        assert is_logged_in_as_admin_teacher(teacher1.new_user)

        c.logout()

        c.login(username=email2, password=password2)

        assert not is_logged_in_as_admin_teacher(teacher2.new_user)

        c.logout()

    @patch("common.helpers.emails.send_dotdigital_email")
    def test_registrations_increment_data(
        self, mock_send_dotdigital_email: Mock
    ):
        c = Client()

        total_activity = TotalActivity.objects.get(id=1)
        teacher_registration_count = total_activity.teacher_registrations
        student_registration_count = total_activity.student_registrations
        independent_registration_count = (
            total_activity.independent_registrations
        )

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-consent_ticked": "on",
                "teacher_signup-teacher_password": "$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn",
                "teacher_signup-teacher_confirm_password": "$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn",
                "g-recaptcha-response": "something",
            },
        )

        assert response.status_code == 302
        mock_send_dotdigital_email.assert_called_once()

        total_activity = TotalActivity.objects.get(id=1)

        assert (
            total_activity.teacher_registrations
            == teacher_registration_count + 1
        )

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": 7,
                "independent_student_signup-date_of_birth_month": 10,
                "independent_student_signup-date_of_birth_year": 1997,
                "independent_student_signup-name": "Test Name",
                "independent_student_signup-email": "test@indy-email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn",
                "independent_student_signup-confirm_password": "$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn$RFVBGT%6yhn",
                "g-recaptcha-response": "something",
            },
        )

        assert response.status_code == 302
        mock_send_dotdigital_email.assert_called()

        total_activity = TotalActivity.objects.get(id=1)

        assert (
            total_activity.independent_registrations
            == independent_registration_count + 1
        )

        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        _, _, access_code = create_class_directly(teacher_email)

        c.login(username=teacher_email, password=teacher_password)
        c.post(
            reverse("view_class", kwargs={"access_code": access_code}),
            {"names": "Student 1, Student 2, Student 3"},
        )

        assert response.status_code == 302

        total_activity = TotalActivity.objects.get(id=1)

        assert (
            total_activity.student_registrations
            == student_registration_count + 3
        )


# CRON view tests


class CronTestClient(APIClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, HTTP_X_APPENGINE_CRON="true")

    def generic(
        self,
        method,
        path,
        data="",
        content_type="application/octet-stream",
        secure=False,
        **extra,
    ):
        wsgi_response = super().generic(
            method, path, data, content_type, secure, **extra
        )
        assert (
            200 <= wsgi_response.status_code < 300
        ), f"Response has error status code: {wsgi_response.status_code}"

        return wsgi_response


class CronTestCase(APITestCase):
    client_class = CronTestClient


class TestUser(CronTestCase):
    # TODO: use fixtures
    def setUp(self):
        teacher_email, _ = signup_teacher_directly(preverified=False)
        create_organisation_directly(teacher_email)
        _, _, access_code = create_class_directly(teacher_email)
        _, _, student = create_school_student_directly(access_code)
        indy_email, _, _ = create_independent_student_directly()

        self.teacher_user = User.objects.get(email=teacher_email)
        self.teacher_user_profile = UserProfile.objects.get(
            user=self.teacher_user
        )

        self.indy_user = User.objects.get(email=indy_email)
        self.indy_user_profile = UserProfile.objects.get(user=self.indy_user)

        self.student_user: User = student.new_user

    @patch("portal.views.cron.user.send_dotdigital_email")
    def send_verify_email_reminder(
        self,
        days: int,
        is_verified: bool,
        view_name: str,
        assert_called: bool,
        mock_send_dotdigital_email: Mock,
    ):
        self.teacher_user.date_joined = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.teacher_user.save()
        self.student_user.date_joined = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.student_user.save()
        self.indy_user.date_joined = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.indy_user.save()

        self.teacher_user_profile.is_verified = is_verified
        self.teacher_user_profile.save()
        self.indy_user_profile.is_verified = is_verified
        self.indy_user_profile.save()

        self.client.get(reverse(view_name))

        if assert_called:
            mock_send_dotdigital_email.assert_any_call(
                ANY, [self.teacher_user.email], personalization_values=ANY
            )

            mock_send_dotdigital_email.assert_any_call(
                ANY, [self.indy_user.email], personalization_values=ANY
            )

            # Check only two emails are sent - the student should never be included.
            assert mock_send_dotdigital_email.call_count == 2
        else:
            mock_send_dotdigital_email.assert_not_called()

        mock_send_dotdigital_email.reset_mock()

    def test_first_verify_email_reminder_view(self):
        self.send_verify_email_reminder(
            6, False, "first-verify-email-reminder", False
        )
        self.send_verify_email_reminder(
            7, False, "first-verify-email-reminder", True
        )
        self.send_verify_email_reminder(
            7, True, "first-verify-email-reminder", False
        )
        self.send_verify_email_reminder(
            8, False, "first-verify-email-reminder", False
        )

    def test_second_verify_email_reminder_view(self):
        self.send_verify_email_reminder(
            13, False, "second-verify-email-reminder", False
        )
        self.send_verify_email_reminder(
            14, False, "second-verify-email-reminder", True
        )
        self.send_verify_email_reminder(
            14, True, "second-verify-email-reminder", False
        )
        self.send_verify_email_reminder(
            15, False, "second-verify-email-reminder", False
        )

    def test_anonymise_unverified_accounts_view(self):
        now = timezone.now()

        for user in [self.teacher_user, self.indy_user, self.student_user]:
            user.date_joined = now - timedelta(
                days=USER_DELETE_UNVERIFIED_ACCOUNT_DAYS + 1
            )
            user.save()

        for user_profile in [self.teacher_user_profile, self.indy_user_profile]:
            user_profile.is_verified = True
            user_profile.save()

        def anonymise_unverified_users(
            days: int,
            is_verified: bool,
            assert_active: bool,
        ):
            date_joined = now - timedelta(days=days, hours=12)

            # Create teacher.
            teacher_user = User.objects.create(
                first_name="Unverified",
                last_name="Teacher",
                username="unverified.teacher@codeforlife.com",
                email="unverified.teacher@codeforlife.com",
                date_joined=date_joined,
            )
            teacher_user_profile = UserProfile.objects.create(
                user=teacher_user,
                is_verified=is_verified,
            )
            Teacher.objects.create(
                user=teacher_user_profile,
                new_user=teacher_user,
                school=self.teacher_user.new_teacher.school,
            )

            # Create dependent student.
            student_user = User.objects.create(
                first_name="Unverified",
                last_name="DependentStudent",
                username="UnverifiedDependentStudent",
                date_joined=date_joined,
            )
            student_user_profile = UserProfile.objects.create(
                user=student_user,
            )
            Student.objects.create(
                user=student_user_profile,
                new_user=student_user,
                class_field=self.student_user.new_student.class_field,
            )

            # Create independent student.
            indy_user = User.objects.create(
                first_name="Unverified",
                last_name="IndependentStudent",
                username="unverified.independentstudent@codeforlife.com",
                email="unverified.independentstudent@codeforlife.com",
                date_joined=date_joined,
            )
            indy_user_profile = UserProfile.objects.create(
                user=indy_user,
                is_verified=is_verified,
            )
            Student.objects.create(
                user=indy_user_profile,
                new_user=indy_user,
            )

            activity_today = DailyActivity.objects.get_or_create(
                date=datetime.now().date()
            )[0]
            daily_teacher_count = activity_today.anonymised_unverified_teachers
            daily_indy_count = activity_today.anonymised_unverified_independents

            total_activity = TotalActivity.objects.get(id=1)
            total_teacher_count = total_activity.anonymised_unverified_teachers
            total_indy_count = total_activity.anonymised_unverified_independents

            self.client.get(reverse("anonymise-unverified-accounts"))

            # Assert the verified users exist
            assert User.objects.get(id=self.teacher_user.id).is_active
            assert User.objects.get(id=self.student_user.id).is_active
            assert User.objects.get(id=self.indy_user.id).is_active

            teacher_user_active = User.objects.get(id=teacher_user.id).is_active
            indy_user_active = User.objects.get(id=indy_user.id).is_active
            student_user_active = User.objects.get(id=student_user.id).is_active

            assert teacher_user_active == assert_active
            assert indy_user_active == assert_active
            assert student_user_active

            activity_today = DailyActivity.objects.get_or_create(
                date=datetime.now().date()
            )[0]
            total_activity = TotalActivity.objects.get(id=1)

            if not teacher_user_active:
                assert (
                    activity_today.anonymised_unverified_teachers
                    == daily_teacher_count + 1
                )
                assert (
                    total_activity.anonymised_unverified_teachers
                    == total_teacher_count + 1
                )

            if not indy_user_active:
                assert (
                    activity_today.anonymised_unverified_independents
                    == daily_indy_count + 1
                )
                assert (
                    total_activity.anonymised_unverified_independents
                    == total_indy_count + 1
                )

            teacher_user.delete()
            indy_user.delete()
            student_user.delete()

        anonymise_unverified_users(
            days=USER_DELETE_UNVERIFIED_ACCOUNT_DAYS - 1,
            is_verified=False,
            assert_active=True,
        )
        anonymise_unverified_users(
            days=USER_DELETE_UNVERIFIED_ACCOUNT_DAYS,
            is_verified=False,
            assert_active=False,
        )
        anonymise_unverified_users(
            days=USER_DELETE_UNVERIFIED_ACCOUNT_DAYS,
            is_verified=True,
            assert_active=True,
        )
        anonymise_unverified_users(
            days=USER_DELETE_UNVERIFIED_ACCOUNT_DAYS + 1,
            is_verified=False,
            assert_active=False,
        )

    @patch("portal.views.cron.user.send_dotdigital_email")
    def send_inactivity_reminder(
        self,
        days: int,
        view_name: str,
        assert_called: bool,
        campaign_name: str,
        mock_send_dotdigital_email: Mock,
        personalization_values=None,
    ):
        self.teacher_user.date_joined = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.teacher_user.save()
        self.student_user.date_joined = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.student_user.save()
        self.indy_user.last_login = timezone.now() - timedelta(
            days=days, hours=12
        )
        self.indy_user.save()

        self.client.get(reverse(view_name))

        if assert_called:
            if personalization_values is not None:
                mock_send_dotdigital_email.assert_any_call(
                    campaign_ids[campaign_name],
                    [self.teacher_user.email],
                    personalization_values=personalization_values,
                )

                mock_send_dotdigital_email.assert_any_call(
                    campaign_ids[campaign_name],
                    [self.indy_user.email],
                    personalization_values=personalization_values,
                )
            else:
                mock_send_dotdigital_email.assert_any_call(
                    campaign_ids[campaign_name],
                    [self.teacher_user.email],
                )

                mock_send_dotdigital_email.assert_any_call(
                    campaign_ids[campaign_name],
                    [self.indy_user.email],
                )

            # Check only two emails are sent - the student should never be included.
            assert mock_send_dotdigital_email.call_count == 2
        else:
            mock_send_dotdigital_email.assert_not_called()

        mock_send_dotdigital_email.reset_mock()

    def test_first_inactivity_reminder_view(self):
        self.send_inactivity_reminder(
            729,
            "first-inactivity-reminder",
            False,
            "inactive_users_on_website_first_reminder",
        )
        self.send_inactivity_reminder(
            730,
            "first-inactivity-reminder",
            True,
            "inactive_users_on_website_first_reminder",
        )
        self.send_inactivity_reminder(
            731,
            "first-inactivity-reminder",
            False,
            "inactive_users_on_website_first_reminder",
        )

    def test_second_inactivity_reminder_view(self):
        self.send_inactivity_reminder(
            972,
            "second-inactivity-reminder",
            False,
            "inactive_users_on_website_second_reminder",
        )
        self.send_inactivity_reminder(
            973,
            "second-inactivity-reminder",
            True,
            "inactive_users_on_website_second_reminder",
            personalization_values={
                "DAYS_LEFT": USER_RETENTION_PERIOD
                - USER_2ND_INACTIVE_REMINDER_DAYS
            },
        )
        self.send_inactivity_reminder(
            974,
            "second-inactivity-reminder",
            False,
            "inactive_users_on_website_second_reminder",
        )

    def test_final_inactivity_reminder_view(self):
        self.send_inactivity_reminder(
            1064,
            "final-inactivity-reminder",
            False,
            "inactive_users_on_website_final_reminder",
        )
        self.send_inactivity_reminder(
            1065,
            "final-inactivity-reminder",
            True,
            "inactive_users_on_website_final_reminder",
            personalization_values={
                "DAYS_LEFT": USER_RETENTION_PERIOD
                - USER_FINAL_INACTIVE_REMINDER_DAYS
            },
        )
        self.send_inactivity_reminder(
            1066,
            "final-inactivity-reminder",
            False,
            "inactive_users_on_website_final_reminder",
        )
